/*
   Class Name:

      CPBuffer.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class represents a P Buffer that can be rendered to.
*/


#include"CPBuffer.h"


// PBuffer functions.
PFNWGLGETEXTENSIONSSTRINGARBPROC wglGetExtensionsStringARB = NULL;

PFNWGLCREATEPBUFFERARBPROC wglCreatePbufferARB = NULL;
PFNWGLGETPBUFFERDCARBPROC wglGetPbufferDCARB = NULL;
PFNWGLRELEASEPBUFFERDCARBPROC wglReleasePbufferDCARB = NULL;
PFNWGLDESTROYPBUFFERARBPROC wglDestroyPbufferARB = NULL;
PFNWGLQUERYPBUFFERARBPROC wglQueryPbufferARB = NULL;

PFNWGLGETPIXELFORMATATTRIBIVARBPROC wglGetPixelFormatAttribivARB = NULL;
PFNWGLGETPIXELFORMATATTRIBFVARBPROC wglGetPixelFormatAttribfvARB = NULL;
PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = NULL;

// WGL_ARB_render_texture
PFNWGLBINDTEXIMAGEARBPROC wglBindTexImageARB = NULL;
PFNWGLRELEASETEXIMAGEARBPROC wglReleaseTexImageARB = NULL;


CPBuffer::CPBuffer() : width(0), height(0), hDC(0), hRC(0), pBufferHandle(0), ID(-1)
{
}


CPBuffer::~CPBuffer()
{
   Shutdown();
}


bool CPBuffer::Initialize(int w, int h, const int *iAttriList,
                          const float *fAttriList, const int *flags)
{
   // Release anything that was here previously.
   Shutdown();
   
   // Check and make sure the right extensions are supported.  If so we continue.
   if(!WGLExtensionSupported("WGL_ARB_pixel_format") ||
      !WGLExtensionSupported("WGL_ARB_pbuffer"))
   {
      return false;
   }

   // These will hold the pixel format, number of formats, and temp HDC.
   int pixelFormat = 0;
   unsigned int numFormats = 0;
   HDC tempCurrenthDC = 0;

   // Save the current device context into a temp HDC variable.
   tempCurrenthDC = wglGetCurrentDC();

   // Error checking.
   if(!tempCurrenthDC) return false;

   // Grab the pixel format.  Send the current device context, int attribute list, float
   // attribute list, max formats, variable to grab the pixel format, and a variable for
   // the number of formats.
   if(!wglChoosePixelFormatARB(tempCurrenthDC, iAttriList, fAttriList, 1,
                               &pixelFormat, &numFormats))
      {
         return false;
      }

   if(numFormats <= 0) return false;

   // Create the p buffer by calling wglCreatePbufferARB().  Send the current device
   // context, pixel format we want, width and height of the buffer, and the flags.
   pBufferHandle = wglCreatePbufferARB(tempCurrenthDC, pixelFormat, w, h, flags);
   
   // Error checking.
   if(!pBufferHandle) return false;

   // Save the device context of the p buffer into our member variable hDC.
   hDC = wglGetPbufferDCARB(pBufferHandle);
   
   if(!hDC) return false;

   // Save the rendering context to the member variable hRC.
   hRC = wglCreateContext(hDC);
   
   if(!hRC) return false;

   // Grab the actual width and height of the p buffer and store in member variables.
   wglQueryPbufferARB(pBufferHandle, WGL_PBUFFER_WIDTH_ARB, &width);
   wglQueryPbufferARB(pBufferHandle, WGL_PBUFFER_HEIGHT_ARB, &height);
	
   return true;
}


bool CPBuffer::MakeCurrent()
{
   // If both the hDC and hRC exists then it might be safe to contine.
   if(!hDC || !hRC) return false;
   
   if(!wglMakeCurrent(hDC, hRC)) return false;

   return true;
}


void CPBuffer::Shutdown(void)
{
   // Release all resources.
   if(hRC)
      {
         wglDeleteContext(hRC);
         hRC = 0;
      }

   if(hDC)
      {
         wglReleasePbufferDCARB(pBufferHandle, hDC);
         hDC = 0;
      }

   wglDestroyPbufferARB(pBufferHandle);
   pBufferHandle = 0;
   width = height = 0;
}


bool ExtensionSupported(char *ext)
{
   char *extension = (char*)glGetString(GL_EXTENSIONS);

   if(strstr(extension, ext) == NULL)
      {
         return false;
      }

   return true;
}


bool WGLExtensionSupported(char *ext)
{
   if(wglGetExtensionsStringARB == NULL)
      wglGetExtensionsStringARB = (PFNWGLGETEXTENSIONSSTRINGARBPROC)wglGetProcAddress("wglGetExtensionsStringARB");

	char *extension = NULL;
	
	if(wglGetExtensionsStringARB)
		extension = (char*)wglGetExtensionsStringARB(wglGetCurrentDC());
	else
		return false;

   if(strstr(extension, ext) == NULL)
      {
         return false;
      }

   return true;
}


bool SetupPBufferARB()
{
	if(!WGLExtensionSupported("WGL_ARB_pbuffer"))
	   {
		   return false;
	   }
	else
	   {
		   wglCreatePbufferARB = (PFNWGLCREATEPBUFFERARBPROC)wglGetProcAddress("wglCreatePbufferARB");
		   wglGetPbufferDCARB = (PFNWGLGETPBUFFERDCARBPROC)wglGetProcAddress("wglGetPbufferDCARB");
		   wglReleasePbufferDCARB = (PFNWGLRELEASEPBUFFERDCARBPROC)wglGetProcAddress("wglReleasePbufferDCARB");
		   wglDestroyPbufferARB = (PFNWGLDESTROYPBUFFERARBPROC)wglGetProcAddress("wglDestroyPbufferARB");
		   wglQueryPbufferARB = (PFNWGLQUERYPBUFFERARBPROC)wglGetProcAddress("wglQueryPbufferARB");

		   if(!wglCreatePbufferARB || !wglGetPbufferDCARB || !wglReleasePbufferDCARB ||
			   !wglDestroyPbufferARB || !wglQueryPbufferARB )
		      {
			      return false;
		      }
	   }


	if(!WGLExtensionSupported("WGL_ARB_pixel_format"))
	   {
		   return false;
	   }
	else
	   {
		   wglGetPixelFormatAttribivARB = (PFNWGLGETPIXELFORMATATTRIBIVARBPROC)wglGetProcAddress("wglGetPixelFormatAttribivARB");
		   wglGetPixelFormatAttribfvARB = (PFNWGLGETPIXELFORMATATTRIBFVARBPROC)wglGetProcAddress("wglGetPixelFormatAttribfvARB");
		   wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC)wglGetProcAddress("wglChoosePixelFormatARB");

		   if(!wglGetExtensionsStringARB || !wglCreatePbufferARB || !wglGetPbufferDCARB)
		      {
			      return false;
		      }
	   }

   return true;
}


bool SetupRenderTexARB()
{
   if(!WGLExtensionSupported("WGL_ARB_render_texture")) return false;
   
   wglBindTexImageARB = (PFNWGLBINDTEXIMAGEARBPROC)wglGetProcAddress("wglBindTexImageARB");
   wglReleaseTexImageARB = (PFNWGLRELEASETEXIMAGEARBPROC)wglGetProcAddress("wglReleaseTexImageARB");

   if(!wglBindTexImageARB || !wglReleaseTexImageARB) return false;

   return true;
}


// Copyright December 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com