/*
   Offscreen Rendering.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform off screen rendering in OpenGL.  To do this
   we will create a off screen surface using a OpenGL P Buffer and create a texture
   out of the scene we render to it.  This texture is then used as a texture map
   when we render a simple quad to the screen.
*/


// Demo: Offscreen Rendering.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN


// All the necessary includes...
#include<windows.h>
#include<gl/gl.h>
#include<gl/glu.h>
#include"CTGALoader.h"
#include"CPBuffer.h"


// Function Prototypes
bool InitializeGL();
void RenderScene();
void Shutdown();


// Global device context.
HDC g_HDC;
HGLRC g_HRC;

// xRotation and yRotation is used to rotate the entire scene.
float xRotation = 0.0f;
float yRotation = 0.0f;

// Object's texture image.
CTGALoader decal;

// PBuffer data.
CPBuffer pBuffer;
int pbWidth = 640;
int pbHeight = 480;


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;                      // Rendering context.
   static HDC hDC;                        // Device context.
   int width, height;                     // The window width and height.
   static POINT oldMousePos;              // Last mouse position.
   static POINT currentMousePos;          // Current mouse position.
   static bool isMouseActive;             // Is the left mouse button down.

   switch(message)
      {
         case WM_CREATE:                  // Windows creation.
            hDC = GetDC(hwnd);            // This gets the device context for our window.
            g_HDC = hDC;                  // Assigns the global device context to this one.
            SetupPixelFormat(hDC);        // Call the pixel format function.

            hRC = wglCreateContext(hDC);  // Creates the rendering context.
            g_HRC = hRC;                  // Assign to global HGLRC.
            wglMakeCurrent(hDC, hRC);     // Makes the rendering context.
            return 0;
            break;

         case WM_CLOSE:                   // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);        // Deletes the rendering context.

            PostQuitMessage(0);           // Says close the program.
            return 0;
            break;

         case WM_SIZE:                    // re-size message.
            height = HIWORD(lParam);      // This gets the height of the window.
            width = LOWORD(lParam);       // This gets the width of the window.

            if(height == 0)               // we don't want it to be possible for a
               {                          // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height);// Resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);    // Sets the projection matrix.
            glLoadIdentity();               // Reset the modelview matrix.

            // Calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);     // Sets the projection matrix.
            glLoadIdentity();               // Reset the modelview matrix.

            return 0;
            break;

         case WM_KEYDOWN:
		      switch(wParam) 
		         {
			         case VK_ESCAPE:
				         PostQuitMessage(0);
				         break;
		         }
            break;

         case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;

			   break;

			case WM_LBUTTONUP:
				isMouseActive = false;

				break;

			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				if(isMouseActive)
					{
                  xRotation -= (currentMousePos.x - oldMousePos.x);
						yRotation -= (currentMousePos.y - oldMousePos.y);
					}

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;

			   break;

         default:
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                        // A message variable.
   WNDCLASSEX windowClass;                         // Your Window class.
   HWND hwnd;                                      // The Window handle.
   bool isFinished;                                // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Handle of application Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// icon when minimized.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,                           // The extended window style.
                        "UltimateGameProgrammingClass",  // window Class name.
                        "Offscreen Rendering - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |   // window style.
                        WS_CLIPSIBLINGS,                 // window style.
                        100, 100,                        // window x, y coordinate.
                        640, 480,                        // window width and height.
                        NULL,                            // handle to parent window.
                        NULL,                            // menu.
                        hInstance,                       // handle to app instance.
                        NULL);                           // pointer to win data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If the initialization of OpenGL fail we can't run the program.
   if(!InitializeGL()) isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.6f, 0.6f, 0.8f, 0.0f);
   glShadeModel(GL_SMOOTH);
   glEnable(GL_DEPTH_TEST);
   glEnable(GL_TEXTURE_2D);

   // Initialize extensions for p buffer.
   if(!SetupPBufferARB()) return false;

   // Attribute list to describe the p buffer.
   int attrList[] =
      {
         WGL_SUPPORT_OPENGL_ARB, true,
         WGL_DRAW_TO_PBUFFER_ARB, true,
         WGL_RED_BITS_ARB, 8,
         WGL_GREEN_BITS_ARB, 8,
         WGL_BLUE_BITS_ARB, 8,
         WGL_ALPHA_BITS_ARB, 8,
         WGL_DEPTH_BITS_ARB, 16,
         WGL_DOUBLE_BUFFER_ARB, false,
         0
      };
   
   pBuffer.Initialize(pbWidth, pbHeight, attrList, NULL, NULL);
   
   glGenTextures(1, &pBuffer.ID);

   glBindTexture(GL_TEXTURE_2D, pBuffer.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, pbWidth, pbHeight, 0,
                GL_RGB, GL_FLOAT, 0);

   pBuffer.MakeCurrent();

      glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
      glShadeModel(GL_SMOOTH);
      glEnable(GL_DEPTH_TEST);
      glEnable(GL_TEXTURE_2D);

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(45.0f, pbWidth / pbHeight, 0.1f, 10.0f);
      
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

   // Share all list with the main window and p buffer (like textures).
   if(!wglShareLists(g_HRC, pBuffer.hRC)) return false;
   if(!wglMakeCurrent(g_HDC, g_HRC)) return false;

   // Load the texture image for our object.
   if(!decal.LoadTGA("decal.tga")) return false;
   glGenTextures(1, &decal.ID);
   
   glBindTexture(GL_TEXTURE_2D, decal.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

   int type = decal.type == 3 ? GL_RGB : GL_RGBA;
   gluBuild2DMipmaps(GL_TEXTURE_2D, type, decal.imageWidth, decal.imageHeight,
                     type, GL_UNSIGNED_BYTE, decal.image);

   return true;
}


void RenderScene()
{
   int flags;
   // Make sure the p buffer is still valid.
   wglQueryPbufferARB(pBuffer.pBufferHandle, WGL_PBUFFER_LOST_ARB, &flags);
   if(flags) return;

   // Render to the texture.
   pBuffer.MakeCurrent();

      // Clear the screen and modelview matrix.
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glLoadIdentity();

      glTranslatef(0.0f, 0.0f, -2.5f);
      glRotatef(-yRotation, 1.0f, 0.0f, 0.0f);
	   glRotatef(-xRotation, 0.0f, 1.0f, 0.0f);

      glBindTexture(GL_TEXTURE_2D, decal.ID);
      glBegin(GL_QUADS);

         glTexCoord2f(0.0f, 0.0f); glVertex3f(-0.5f, -0.5f, 0.0f);
         glTexCoord2f(1.0f, 0.0f); glVertex3f(0.5f, -0.5f, 0.0f);
         glTexCoord2f(1.0f, 1.0f); glVertex3f(0.5f, 0.5f, 0.0f);
         glTexCoord2f(0.0f, 1.0f); glVertex3f(-0.5f, 0.5f, 0.0f);

      glEnd();
      
      // Copy the scene created in the pbuffer to the pbuffer's texture.
      glBindTexture(GL_TEXTURE_2D, pBuffer.ID);
      glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0,
                         pBuffer.width, pBuffer.height);

   // Render a quad with the render texture as a texture after switching
   // back to the main window for rendering.
   if(!wglMakeCurrent(g_HDC, g_HRC)) return;

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   glTranslatef(0.0f, 0.0f, -2.5f);

   glBindTexture(GL_TEXTURE_2D, pBuffer.ID);
   glBegin(GL_QUADS);

      glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, 0.0f);
      glTexCoord2f(1.0f, 0.0f); glVertex3f(1.0f, -1.0f, 0.0f);
      glTexCoord2f(1.0f, 1.0f); glVertex3f(1.0f, 1.0f, 0.0f);
      glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f, 1.0f, 0.0f);

   glEnd();
   
   SwapBuffers(g_HDC);
}


void Shutdown()
{
   // Free texture image.
   decal.FreeImage();
   pBuffer.Shutdown();

   glDeleteTextures(1, &decal.ID);
   glDeleteTextures(1, &pBuffer.ID);
}


// Recap:

// Using p buffers to render a scene to takes a little work but the end
// result is nice.  Once we render the texture to the offscreen buffer
// we place that scene in a texture that can be rendered onto any surface.
// In our case we render the texture to a quad that is twice the size of the
// object we rendered into the p buffer.


// Copyright December 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com